from upyog.imports import *
from upyog.os.read_files import *


@fastcore.patch
def ls(self: Path, recurse: bool = False, extensions: Optional[List[str]] = None):
    return get_files(self, recurse=recurse, extensions=extensions)


@fastcore.patch
def list_dirs(self: Path):
    dirs = []
    for f in os.listdir(self):
        f = Path(os.path.join(self, f))
        if f.is_dir():
            dirs.append(f)
    return dirs



from upyog.imports import *
from upyog.utils.utils import flatten


"""
Taken from fastaiv1: https://github.com/fastai/fastai1/blob/a8327427ad5137c4899a1b4f74745193c9ea5be3/fastai/data_block.py#L22-L47
"""

__all__ = ["PathLike", "get_files", "get_image_files", "get_video_files", "collate_image_filenames"]


def get_files(
    path: PathLike,
    extensions: Collection[str] = None,
    recurse: bool = False,
    exclude: Optional[Collection[str]] = None,
    include: Optional[Collection[str]] = None,
    presort: bool = False,
    followlinks: bool = False,
) -> List[Path]:
    """
    Return list of files in `path` that have a suffix in `extensions`; optionally `recurse`.
    Use `include` and `exclude` for including/excluding folder names, `presort` to sort.
    """
    # fmt: off
    if recurse:
        res = []
        for i,(p,d,f) in enumerate(os.walk(path, followlinks=followlinks)):
            # skip hidden dirs
            if include is not None and i==0:   d[:] = [o for o in d if o in include]
            elif exclude is not None and i==0: d[:] = [o for o in d if o not in exclude]
            else:                              d[:] = [o for o in d if not o.startswith('.')]
            res += _get_files(path, p, f, extensions)
        if presort: res = sorted(res, key=lambda p: _path_to_same_str(p), reverse=False)
        return res
    else:
        f = [o.name for o in os.scandir(path) if o.is_file()]
        res = _get_files(path, path, f, extensions)
        if presort: res = sorted(res, key=lambda p: _path_to_same_str(p), reverse=False)
        return res
    # fmt: on


def _path_to_same_str(p_fn: PathLike) -> str:
    "path -> str, but same on nt+posix, for alpha-sort only"
    s_fn = str(p_fn)
    s_fn = s_fn.replace("\\", ".")
    s_fn = s_fn.replace("/", ".")
    return s_fn


def _get_files(parent, p, f, extensions) -> list:
    p = Path(p)  # .relative_to(parent)
    if isinstance(extensions, str):
        extensions = [extensions]
    low_extensions = [e.lower() for e in extensions] if extensions is not None else None
    res = [
        p / o
        for o in f
        if not o.startswith(".")
        and (extensions is None or f'.{o.split(".")[-1].lower()}' in low_extensions)
    ]
    return res


_IMAGE_EXTENSIONS = set(
    k for k, v in mimetypes.types_map.items() if v.startswith("image/")
)
_VIDEO_EXTENSIONS = set(
    [k for k, v in mimetypes.types_map.items() if v.startswith("video/")] + [".mkv"]
)


def get_image_files(
    path: Union[PathLike, List[PathLike]],
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    recurse: bool = True,
) -> List[Path]:

    if not isinstance(path, list):
        path = [path]

    paths = path  # For backward compatibility
    all_files = []
    for path in paths:
        files = get_files(
            path=path,
            include=include,
            exclude=exclude,
            recurse=recurse,
            extensions=_IMAGE_EXTENSIONS,
        )
        all_files += files

    return flatten(files)


def get_video_files(
    path: PathLike,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    recurse: bool = True,
) -> List[Path]:
    return get_files(
        path=path,
        include=include,
        exclude=exclude,
        recurse=recurse,
        extensions=_VIDEO_EXTENSIONS,
    )


def collate_image_filenames(
    filenames: List[os.PathLike] = None,
    img_folders: Optional[List[os.PathLike]] = None,
    sort: bool = True,
    return_str: bool = False,
) -> List[PathLike]:
    """
    Take a list of `fnames` and `img_folders`, and return a flattened list of
    all image files in `img_folders` and `fnames` as a single list
    """
    all_files = []
    if filenames is not None:
        assert isinstance(filenames, list)
        filenames = [Path(f) for f in filenames]
        all_files.extend(filenames)

    if img_folders is not None:
        if isinstance(img_folders, (str, Path)):
            img_folders = [img_folders]
        for folder in img_folders:
            if not Path(folder).exists():
                raise FileNotFoundError(f"{folder} not found on disk.")
        for folder in img_folders:
            all_files.extend(get_image_files(folder))

    if sort:
        all_files = sorted(all_files)

    if return_str:
          return [str(f) for f in all_files]
    else: return all_files



from upyog.ml.imports import *
from upyog.os import *


__all__ = ["to_np", "to_torch", "save_tensor_to_json", "pick_device"]


# NOTE: PyTorch Lightning specific... do not use unless you know exactly what this
# function is meant for...
def get_num_gpus(gpus: Union[List[int], int]) -> Union[None, int]:
    if is_platform_macos():
        return None
    if gpus == -1:
        return torch.cuda.device_count()
    else:
        assert isinstance(gpus, list)
        return len(gpus)


def to_np(
    data: Union[Tensor, np.ndarray, Collection[Any]],
    dtype=np.float32
):
    "Convert a Tensor / array / collection of numbers to a NP array and optionally change its dtype"
    if   isinstance(data, np.ndarray): pass
    elif isinstance(data, Tensor):     data = data.detach().cpu().numpy()
    else:                              data = np.array(data)

    if dtype is not None: return data.astype(dtype)
    else:                 return data


def to_torch(x) -> Tensor:
    assert isinstance(x, (np.ndarray, torch.Tensor))
    return x.detach() if isinstance(x, torch.Tensor) else torch.from_numpy(x)


def save_tensor_to_json(data: Union[Tensor, np.ndarray], path: PathLike):
    "Save tensor to disk as a JSON file"
    data = to_np(data) if isinstance(data, Tensor) else data
    if not isinstance(data, (Tensor, np.ndarray)):
        raise ValueError(f"Invalid `data` type. Expected tensor/array, got {type(data)}")
    data = data.tolist()

    path = Path(path).with_suffix(".json")
    # path = Path(path)
    # filename = f"{path.stem}.json"  # Ensure extension is .json
    # path = path.parent / filename
    write_json(data, path, indent=None)


def pick_device(device=None) -> torch.device:
    if device is None:
        if torch.cuda.is_available():
            return torch.device("cuda:0")
        else:
            # MacOS
            if platform.system() == "Darwin":
                try: return torch.device("mps")
                except: return torch.device("cpu")
            else:
                return torch.device("cpu")

    if isinstance(device, int):
        return torch.device(f"cuda:{device}")

    if device == "cpu":
        return torch.device("cpu")

    if device == "mps":
        return torch.device("mps")

    if isinstance(device, torch.device):
        return device

    raise ValueError(f"Invalid device value: {device}")



from upyog.ml.imports import *
from upyog.os import get_image_files, collate_image_filenames
from upyog.image import load_image


__all__ = ["ImageInferenceDataset", "image_filename_collate_fn"]


class ImageInferenceDataset(Dataset, ABC):
    """
    Base class that gathers files from a myriad number of folders / lists of
    filenames and stores them as `self.fnames`.

    Subclasses must implement the `__getitem__` method with the appropriate
    preprocessing for the model
    """

    def __init__(
        self,
        fnames: List[os.PathLike] = None,
        img_folders: Optional[List[os.PathLike]] = None,
        sort_fnames: bool = True,
    ):
        self.fnames = collate_image_filenames(fnames, img_folders)
        if sort_fnames:
            self.fnames.sort()

        num_unique_files = len(set(self.fnames))
        if not num_unique_files == len(self.fnames):
            before = len(self.fnames)
            logger.warning(f"Found {before} - {num_unique_files} duplicate filenames")

            self.fnames = sorted(set(self.fnames))
            after = len(self.fnames)

            logger.info(f"Removed duplicate filenames. Total no. filenames: {before} -> {after}")

    def post_init(self):
        pass

    def __len__(self):
        return len(self.fnames)

    @abstractmethod
    def __getitem__(self, i):
        pass

    def __repr__(self):
        info = [
            f"{self.__class__.__name__}(...)",
            f"Num Images: {len(self)}",
        ]
        return "\n".join(info)


def image_filename_collate_fn(items):
    fnames = L(items).itemgot(0)
    imgs = L(items).itemgot(1)

    return (list(fnames), torch.stack(list(imgs)))


def read_image(fn: Union[str, os.PathLike]) -> torch.Tensor:
    img = load_image(fn)
    return TF.to_tensor(img)



from imp import reload
from upyog.imports import *
from annoy import AnnoyIndex
from rich import print

__all__ = ["AnnoySimilarityMetrics", "AnnoyIndexer", "AnnoyDataFrameIndexer"]


class AnnoySimilarityMetrics(Enum):
    Angular = "angular"
    Euclidean = "euclidean"
    Manhattan = "manhattan"
    Hamming = "hamming"
    Dot = "dot"


class AnnoyIndexer:
    """
    Wrapper for Spotify's Annoy
    See https://github.com/spotify/annoy for more details

    Selected Args:
        * `path`: If given, builds the index on disk to this path, and adds
                          the suffix ".ann" to it
    """

    def __init__(
        self,
        dimension: int,
        n_trees: int = 10,
        search_k: int = -1,
        metric: AnnoySimilarityMetrics = AnnoySimilarityMetrics.Euclidean,
        path: PathLike = None,
    ):
        if not isinstance(metric, AnnoySimilarityMetrics):
            raise TypeError(
                f"Metric must be of type {AnnoySimilarityMetrics}, got {type(metric)} instead"
            )

        self.n_trees = n_trees
        self.search_k = search_k
        self.metric = metric.value
        self.dimension = dimension
        self.path = path
        self.indexer = AnnoyIndex(self.dimension, self.metric)

    def build_from_indexed_dataframe(self, df: pd.DataFrame, colname: str):
        "Builds an index from a DataFrame with vectors stored in `colname`"
        start = time.time()

        for (index, vector) in df[[colname]].itertuples(index=True):
            self.indexer.add_item(index, vector)
        self.indexer.build(self.n_trees, n_jobs=-1)

        end = time.time()
        print(f"Took {round(end-start, 2)} seconds to build index")

    def save(self):
        assert self.path
        self.indexer.save(str(self.path))
        print(f"Saved index to {self.path}")

    def load_from_disk(self, mmap: bool = True):
        self.indexer.load(str(self.path), prefault=False if mmap else True)
        print(
            f"Loaded index from disk with {self.indexer.get_n_items()} items ",
            f"and {self.indexer.get_n_trees()} trees",
        )

    def search(self, index: int, num_items: int = 100):
        return self.indexer.get_nns_by_item(index, num_items)

    # def __repr__(self):
    #     return f"AnnoyIndex with {self.indexer.get_n_items()} items and {self.indexer.get_n_trees()} trees"


class AnnoyDataFrameIndexer(AnnoyIndexer):
    def __init__(
        self,
        df: pd.DataFrame,
        vector_column: str,
        index_column: str,
        path: PathLike,
        search_k: int = -1,
        n_trees: int = 10,
        metric: AnnoySimilarityMetrics = AnnoySimilarityMetrics.Euclidean,
    ):
        """Spotify's Annoy Indexer that can be used with a DataFrame

        Parameters
        ----------

        df : pd.DataFrame
        vector_column : str
            name of the column containing arrays to build the index from
        index_column : str
            name of the column that you will be running queries from. Typically
            a filename column. You will be able to search for nearest neighbors
            using `indexer[name]`
        path : PathLike, optional
            Path to save the index to, by default None
        search_k : int, optional
            , by default -1
        n_trees : int, optional
            num. trees for the indexer, by default 10
        metric : AnnoySimilarityMetrics, optional
            Metric to build the index on, by default AnnoySimilarityMetrics.Euclidean"""

        vector = df[vector_column].iloc[0]
        assert isinstance(vector, np.ndarray)

        self.path = path
        self.index = df[index_column]
        self.df = df.reset_index(drop=True)

        super().__init__(len(vector), n_trees, search_k, metric, path)

        if Path(self.path).exists():
            self.load_from_disk()
        else:
            self.build_from_indexed_dataframe(self.df, vector_column)
            self.save()

        # Now, df has a column 'index' which has the annoy `indexer`'s indexes
        self.df = self.df.reset_index(drop=False)
        self.df = self.df.set_index(index_column)

    def search(self, index: str, num_items: int = 100):
        index = self.df.loc[index, "index"]
        idxs: List[int] = super().search(index, num_items)

        return self.df.iloc[idxs]

    def __getitem__(self, index):
        return self.search(index)



#!/usr/bin/env python

## SOURCE ==> https://github.com/fastai/fastai_dev/blob/master/dev_nb/notebook2script.py
## with minor modifications

# fmt: off

import json, re
from pathlib import Path
from upyog.cli import Param as P
from upyog.cli import call_parse

def is_export(cell):
    if cell['cell_type'] != 'code': return False
    src = cell['source']
    if len(src) == 0 or len(src[0]) < 7: return False
    #import pdb; pdb.set_trace()
    return re.match(r'^\s*#\s*export\s*$', src[0], re.IGNORECASE) is not None

@call_parse
def notebook2script(fname: P("Path to the notebook to be converted")):
    fname = Path(fname)
    fname_out = fname.parent/f"{fname.stem}.py"
    print(fname_out)
    main_dic = json.load(open(fname,'r'))
    cells = main_dic['cells']
    code_cells = [c for c in cells if is_export(c)]
    module = f'''
        # Autogenerated from {fname}
        # file to edit: {fname.name}\n\n'''
    for cell in code_cells: module += ''.join(cell['source'][1:]) + '\n\n'
    # remove trailing spaces
    module = re.sub(r' +$', '', module, flags=re.MULTILINE)
    with open(fname_out,'w') as f: f.write(module[:-2])
    print(f"Converted {fname} to {fname_out}")

# fmt: on


from upyog.ml.imports import *


__all__ = ["compare_state_dicts", "set_bn_eval", "load_weights_from_checkpoint"]


ModuleOrWeights = Union[nn.Module, OrderedDict]


def compare_state_dicts(
    m1: ModuleOrWeights, m2: ModuleOrWeights, verbose=True, name: str = "models"
) -> bool:
    mismatch_keys = []
    m1 = m1.state_dict() if isinstance(m1, nn.Module) else m1
    m2 = m2.state_dict() if isinstance(m2, nn.Module) else m2

    for (k1, m1v), (k2, m2v) in zip(m1.items(), m2.items()):
        assert k1 == k2, f"Key mismatch, looks like you're comparing different models"
        if not torch.equal(m1v.cpu(), m2v.cpu()):
            logger.error(
                f"Mismatch in <fg #0F77A9>{name}</> in layer <fg #0F77A9>{k1}</>"
            )
            mismatch_keys.append(k1)

    if not mismatch_keys == []:
        err_msg = f"Weights didn't match exactly in {len(mismatch_keys)} layers: {mismatch_keys}"
        raise RuntimeError(err_msg if verbose else "")
    logger.success(f"`state_dict`s of both <fg #0F77A9>{name} </> match!")
    return True


BN_LAYER_TYPES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)
def set_bn_eval(m: nn.Module) -> None:
    "Set bn layers in eval mode for all recursive children of `m`."
    for l in m.children():
        if isinstance(l, BN_LAYER_TYPES) and not next(l.parameters()).requires_grad:
            l.eval()
        set_bn_eval(l)


def load_weights_from_checkpoint(
    path_checkpoint: os.PathLike,
    model: nn.Module,
    replace_pattern: str = "",
    replace_with: str = "",
    re_count: int = 0,
    strict: bool = True,
    verbose: bool = True,
    device: Union[None, torch.device, str] = None,
):
    """
    Loads the state_dict from `path_checkpoint`, which can be a path on disk or a URL

    * If the checkpoint is an elaborate one, then it checks for weights inside the
      'state_dict' or 'model' key.
    * You can use `replace_pattern` and `replace_with` to regex-replace keys in the
      checkpoint
    * If `strict`, it's ensured that none of the keys in the `model`'s state dict
      are missing from the checkpoint
    * Use `re_count` to limit the number of replacements. Defaults to 0 i.e. replace all
      Use if you want to sub "model.model.0" to "model.0" (re_count=1) instead of "0" (re_count=0)
    """
    device = device or "cpu"
    is_url_ckpt = "http" in path_checkpoint and "://" in path_checkpoint

    if is_url_ckpt:
        pt_wts = torch.hub.load_state_dict_from_url(path_checkpoint, map_location=device)
    else:
        pt_wts = torch.load(path_checkpoint, map_location=device)

    if not isinstance(pt_wts, OrderedDict):
        if not isinstance(pt_wts, dict):
            raise TypeError(f"Expected to load a `dict` checkpoint, got {type(pt_wts)}")

        if   "state_dict" in pt_wts.keys(): wts_key = "state_dict"
        elif "model"      in pt_wts.keys(): wts_key = "model"

        else:
            raise ValueError(f"No appropriate key found for loading model weights")

        pt_wts = pt_wts[wts_key]
        assert isinstance(pt_wts, OrderedDict)

    pt_wts = {
        re.sub(replace_pattern, replace_with, layer_name, count=re_count): weights
        for layer_name, weights in pt_wts.items()
    }

    result = model.load_state_dict(pt_wts, strict=False)
    missing, unexpected = result.missing_keys, result.unexpected_keys

    if verbose:
        print(f"Loaded weights from {path_checkpoint}")
        print(f"Missing Keys In Checkpoint\n: {missing}")
        print()
        print(f"Extra Keys From Checkpoint That Weren't Loaded\n: {unexpected}")
    if strict:
        assert missing == [], f"Missing Keys: {missing}"



from upyog.ml.imports import *


__all__ = ["SoftmaxProcessor", "SigmoidProcessor"]


@dataclass
class Prediction:
    label: List[str]
    confidence: List[float]

    def __post_init__(self):
        if isinstance(self.label,      np.ndarray): self.label = self.label.tolist()
        if isinstance(self.confidence, np.ndarray): self.confidence = self.confidence.tolist()


class ActivationsProcessor(ABC):
    "Post processors for predictions coming out of Softmax / Sigmoid layers"

    @property
    @abstractmethod
    def labels(self) -> np.ndarray:
        ...

    @abstractmethod
    def _forward(self):
        ...

    def _type_check(self, preds: np.ndarray):
        if not isinstance(preds, np.ndarray):
            raise TypeError(f"Expected `preds` to be an array, got {type(preds)}")

    def __call__(self, preds: np.ndarray) -> Union[Prediction, List[Prediction]]:
        self._type_check(preds)

        if self._is_singular_pred(preds): return          self._forward(preds)
        else:                             return list(map(self._forward, preds))

    @staticmethod
    def _is_singular_pred(preds: np.ndarray):
        # assert preds.ndim in [1, 2]
        return preds.ndim == 1

    # @abstractmethod
    # def activate(self, preds: np.ndarray):
    #     ...


class SoftmaxProcessor(ActivationsProcessor):
    def __init__(self, labels: List[str], top_k: int = 1):
        self.top_k = top_k
        self._labels = np.array(labels)

    @property
    def labels(self): return self._labels

    def _forward(self, preds) -> Prediction:
        index = np.argsort(preds)[-self.top_k:]  # argsort gives idxs in ascending order
        confidence = preds[index]
        label = self.labels[index]

        return Prediction(label, confidence)


class SigmoidProcessor(ActivationsProcessor):
    def __init__(self, labels: List[str], thresh: float = 0.8):
        self.thresh = thresh
        self._labels = np.array(labels)

    @property
    def labels(self): return self._labels

    def _forward(self, preds) -> Prediction:
        index = np.where(preds > self.thresh)
        confidence = preds[index]
        label = self.labels[index]

        return Prediction(label, confidence)



from upyog.ml.imports import *
import coremltools as ct


__all__ = ["get_coreml_input", "add_top_level_metadata"]

MLModelSpec = ct.proto.Model_pb2.Model


# TODO: Get from upyog
def get_coreml_input(
    H: int,
    W: int,
    norm_mean: Optional[Tuple[float, float, float]] = (0.485, 0.456, 0.406),
    norm_stdev: Optional[Tuple[float, float, float]] = (0.229, 0.224, 0.225),
    input_name: str = "Image",
    input_format: Literal["RGB", "BGR"] = "RGB",
) -> List[ct.ImageType]:
    """
    Get CoreML input image type to feed into a model. If using `norm_mean` and
    `norm_stdev`, it does preprocessing as described (ImageNet example) here:
    https://github.com/apple/coremltools/issues/1160

    If `norm_mean` or `norm_stdev` is None, inputs are are not normalised
    and are in the range of 0-255 (as is the case for YOLOX)

    Returns a list so you can call the function inside the converter alone
        `ct.convert(model, inputs = get_coreml_input(...))`
    """
    if norm_mean and norm_stdev:
        assert len(norm_mean) == 3 and len(norm_stdev) == 3
        scale_div = sum(norm_stdev) / len(norm_stdev)
        scale = 1.0 / (255.0 * scale_div)
        bias = list(map(lambda x: -x / scale_div, norm_mean))
    else:
        scale = 1.0
        bias = None

    input_type = ct.ImageType(
        input_name,
        shape=torch.rand(1, 3, H, W).shape,
        scale=scale,
        bias=bias,
        color_layout=input_format,
    )
    return [input_type]


def add_top_level_metadata(
    mlmod: Union[ct.models.MLModel, MLModelSpec],
    version: str,
    author: str,
    license: str,
    description: str,
) -> ct.models.MLModel:
    ""

    """
    NOTE: We always use the model spec to add the metadata because editing an .mlpackage
    file directly is unreliable -- the metadata shows up in the model in an interactive
    python session but gets erased when you save the model file. Wtf??
    """

    if isinstance(mlmod, MLModelSpec):
        pass

    elif isinstance(mlmod, ct.models.MLModel):
        spec = mlmod.get_spec()

    spec.description.metadata.author = author
    spec.description.metadata.license = license
    spec.description.metadata.shortDescription = description

    check_version_string(version)
    spec.description.metadata.versionString = version

    try:
        return ct.models.MLModel(spec)
    except:
        return ct.models.MLModel(spec, weights_dir=mlmod.weights_dir)


def check_version_string(version: str):
    if not len(version.split(".")) == 3:
        raise ValueError(
            f"Version string seems to be in the wrong format. Expected something like "
            f"'X.Y.Z' or 'X.Y.Z-rc3' ( <MAJOR.MINOR.PATCH-EXTRA> ), got {version} instead. "
            f"See https://semver.org/ for more info"
        )



"""
SOURCE: https://github.com/khuyentran1401/rich-dataframe/blob/master/rich_dataframe/rich_dataframe.py
Some minor modifications made re. color, caption, etc
"""

from upyog.imports import *


def cast_to_df(items: Union[pd.DataFrame, pd.Series, dict]):
    if   isinstance(items, pd.Series):    return pd.DataFrame(items)
    elif isinstance(items, dict):         return pd.DataFrame(items.items())
    elif isinstance(items, pd.DataFrame): return items
    else:
        raise TypeError(f"Expected `items` to be `pd.DataFrame` |  `pd.Series` | `dict`, got {type(items)} instead")

def print_df(
    df: Union[pd.DataFrame, pd.Series],
    title: Optional[str] = None,
    headers = "keys", tablefmt = "psql",
    **tabulate_kwargs,
):
    from tabulate import tabulate

    if title: print(title)
    df = cast_to_df(df)

    table = tabulate(
        df,
        headers = headers,
        tablefmt = tablefmt,
        **tabulate_kwargs
    )
    print(table)

    return table

def print_df(df: Union[pd.DataFrame, pd.Series, dict]):
    from prettytable import PrettyTable

    df = cast_to_df(df)

    table = PrettyTable(max_width=50)
    table.field_names = df.columns  # Set the column names

    # Add the rows to the table
    for row in df.itertuples(index=False):
        table.add_row(row)

    print(table)



from upyog.imports import *
from upyog.utils import *
from upyog.os import *
from upyog.cli import *
from upyog.image import *

logger.opt(colors=True)
tqdm.pandas()
ImageFile.LOAD_TRUNCATED_IMAGES = True

# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def in_jupyter_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


if not in_jupyter_notebook():
    install()



import argparse
import enum
import functools
import importlib
import io
import inspect
import json
import math
import mimetypes
import operator
import os
import platform
import random
import re
import shutil
import socket
import sys
import time
import typing
import uuid
import warnings
import subprocess
import gc

from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import partial, reduce
from pathlib import Path
from pprint import pprint
from types import SimpleNamespace, MethodType
from typing import *
from collections import OrderedDict, defaultdict, namedtuple
from abc import abstractmethod, ABC

import fastcore.all as fastcore
import numpy as np
import pandas as pd
import PIL
import pyperclip
import rich

from fastcore.all import L
from loguru import logger
from PIL import Image, ImageDraw, ImageFile, ImageFont
from PIL import features as PILFeatures
from rich.progress import track
from rich.traceback import install
from tqdm.auto import tqdm
from typing_extensions import Literal

ImageFile.LOAD_TRUNCATED_IMAGES = True
PathLike = Union[str, Path]

logger = logger.opt(colors=True)

def is_package_available(name: str) -> bool:
    if importlib.util.find_spec(name):
          return True
    else: return False



from .imports import *
from .activations import *
from .eval_dataset import *
from .model_utils import *
from .preprocessing import *
from .utils import *
from .visualise import *


if IS_COREML_AVAILABLE:
    from .coreml_utils import *



from upyog.imports import *
from upyog.image import *
from upyog.os import *
from upyog.utils import *
from upyog.cli import *
from upyog.cli import Param as P


@call_parse
def create_image_grid_from_folders(
    # fmt: off
    i: P("An arbitrary number of inputs folders", str, nargs="+") = None,
    o: P("Output path - must be a folder (created if it doesn't exist)", str) = None,
    H: P("Height of _each_ image inside the grid", int) = 384,
    W: P("Width of each image inside the grid", int) = 640,
    N: P("(Optional) Max no. of images to select. If not specified, all are selected", int) = None,
    num_columns: P("Number of columns in the grid", int) = 3,
    sample:      P("Random sample this no. of files from the folder", int) = None,
    shuffle:   P("(Bool) shuffle the order of files", action="store_true") = False,
    foreach:     P("(Bool) Create a dedicated grid for _each_ input folder given", action="store_true") = False,
    exp_quality: P("Export quality (0-100). ~85 is a good tradeoff", int) = 85,
    # fmt: on
):
    """
            --------- IMAGE GRID CREATOR ---------

    Create image grids from multiple folders. By default, files from
    all folders are combined to create a single grid. You can pick a
    random sample of `N` images  using the `sample` arg, or shuffle
    the order of images by passing `--shuffle`.

    To create grids for multiple folders, use the `--foreach` flag.

     Usage
    -------

    generate-image-grid                 \\
        --i inp-folder-1  inp-folder-2  \\
        --o ~/Desktop                   \\
        --sample 200      # Select 200 randomly sampled images

    generate-image-grid                 \\
        --i inp-folder-1  inp-folder-2  \\
        --o ~/Desktop                   \\
        --shuffle                       \\
        --foreach         # Creates 2 grids, one for each input folder
    """
    input_paths = []
    output_path = Path(o)
    for folder in i:
        path = Path(folder)
        assert path.is_dir(), f"!! Given input '{path}' is not a folder !!"
        input_paths += [path]

    # Output directory validation
    if not output_path.suffix == "":
        raise ValueError(
            f"Expected a folder path, but looks like you entered a filepath instead: {output_path}"
        )
    output_path.mkdir(exist_ok=True)

    # Create grids
    if foreach:
        for path in input_paths:
            files = get_image_files(path)
            grid = make_grid_from_files(files, num_columns, (W, H), sample, shuffle, N)

            export_path = output_path / f"{path.stem}.jpg"
            grid.save(export_path, quality=exp_quality)
            rich.print(f"Saved to {export_path}")
    else:
        files = flatten(get_image_files(path) for path in input_paths)
        grid = make_grid_from_files(files, num_columns, (W, H), sample, shuffle, N)

        export_path = output_path / "Image-Grid.jpg"
        grid.save(export_path, quality=exp_quality)
        rich.print(f"Saved to {export_path}")


def make_grid_from_files(files, ncol, size_WH, sample=None, shuffle=False, N=None):
    if sample:  files = random.sample(files, sample)
    if shuffle: random.shuffle(files)
    if N:       files = files[:N]

    imgs = [load_image(fn) for fn in files]
    return make_img_grid(imgs, ncol, size_WH, pad=True)



from upyog.imports import *
from upyog.os.read_files import *
from upyog.utils import *
from upyog.cli import *


__all__ = ["move_files", "cleanup_duplicates"]


@call_parse
def remove_duplicates_from_folder(
    i: P("Single input folder that is to be cleaned up", str) = None,
    t: P(
        "An arbitrary number of target folders that contain duplicate files that are to be removed from the input folder `--i`",
        str,
        nargs="+",
    ) = None,
):
    """
            --------- DUPLICATE REMOVER ---------

    This program scans the _names_ of the files in the input and target folders
    recursively, and removes all the files from the `input` folder that exist in
    `target` folders

    If input folder has the following files:
    ├── input-folder
    │   ├── file1.jpg
    │   ├── file2.jpg
    │   └── file3.jpg

    And the target folder(s) have the following files:
    ├── target-folder-1
    │   ├── file1.jpg
    │   ├── other-random-file-1.jpg

    ├── target-folder-2
    │   ├── file2.jpg
    │   ├── other-random-file-2.jpg

    Using this program would remove `file1.jpg` and `file2.jpg` from the input folders

    Usage:
    ------

    remove-duplicates-from-folder
        --i input-folder
        --t target-folder-1   target-folder-2   ...
    """
    cleanup_path = Path(i)
    target_paths = [Path(p) for p in t]

    target_fnames = flatten([get_files(path, recurse=True) for path in target_paths])
    target_fnames = [f.stem for f in target_fnames]

    cleanup_files = get_files(cleanup_path, recurse=True)
    cleanup_fnames = [f.stem for f in cleanup_files]

    fnames_to_delete = set.intersection(set(cleanup_fnames), set(target_fnames))

    counter = 0
    for f in track(cleanup_files, "Scanning Files..."):
        if f.stem in fnames_to_delete:
            print(f"Deleting {f.name}")
            counter += 1
            f.unlink()

    print(f"Deleted {counter} files")


@call_parse
def add_parent_folder_name(
    # i: P("Input folder(s) and files", str, nargs="+"),
    i: P("Input folder", str),
    sep: P("Separator", str) = "_",
):
    files = get_files(i, recurse=True)
    for file in tqdm(files):
        if file.is_file():
            new_name = file.with_name(f"{file.parent.name}{sep}{file.name}")
            print(f"Renamed: {file.name} => {new_name}")
            file.rename(new_name)


@call_parse
def find_common_files_between_folders(
    i: P("Input folders", str, nargs="+"),
    include_parent: P("Include name of parent folder?", bool) = False,
):
    """
            --------- COMMON FILES DETECTOR ---------

    This program scans an arbitray number of input folders and prints the
    number of common elements between these folders.

    Say we have the following input folders:
    ├── input-folder-1
    │   ├── file1.jpg
    │   ├── file2.jpg
    │   └── file3.jpg
    ├
    ├── input-folder-2
    │   ├── file3.jpg
    │   ├── file4.jpg

    Where there's one common file between input-folder-1 and 2, the result
    is as follows:

        {
            "input-folder-1 (3)": {
                "input-folder-2 (2)": 1
            },
            "input-folder-2 (2)": {
                "input-folder-1 (3)": 1
            }
        }

    Note that the number of files in each folder is added in brackets.

    Usage:
    ------

    find-common-files-between-folders  input-folder-1 input-folder-2

    """
    assert i
    input_folders = i

    def folder_with_parent(f, files=None):
        f = Path(f)
        result = f"{f.parent.name} / {f.name}" if include_parent else f.name
        return f"{result} ({len(files)})" if files is not None else result

    folder2files = {folder: get_image_files(folder) for folder in input_folders}
    folder2files = {
        folder: [f.name for f in files] for folder, files in folder2files.items()
    }
    folder2files = {folder_with_parent(k, v): v for k, v in folder2files.items()}

    # folder_names = [folder_with_parent(f) for f in folder2files.keys()]
    # results = dict.fromkeys(folder_names)
    results = dict.fromkeys(folder2files)
    for f1 in folder2files.keys():
        f1_common = {}
        for f2, files in folder2files.items():
            if f2 == f1:
                continue

            common = set(folder2files[f1]).intersection(files)
            f1_common[f2] = len(common)

        results[f1] = f1_common

    rich.print_json(data=results)
    return results


@call_parse
def print_folder_distribution(
    parent_folders: P("Parent folders", str, nargs="+")
) -> Dict[str, pd.DataFrame]:
    """
    Prints the number of files inside each given `parent_folders`.
    """
    results = {}

    for folder in parent_folders:
        freq = {}
        folder = Path(folder)
        for subdir in sorted(folder.list_dirs()):
            freq[subdir.name] = len(get_files(subdir, recurse=True))

        freq = pd.DataFrame([freq]).T.reset_index()
        freq.columns = ["Sub Folders", "Frequency"]
        rich.print()
        rich.print(f"---------- {folder.name.upper()} ----------")
        freq = print_df(freq)

        results[folder.name] = freq

    return results


@call_parse
def move_files(
    i: P(help="Path to the input directories", type=str, nargs="+") = None,
    o: P(help="Path to the output directory", type=str) = None,
    move: P(help="Move files (destructively)", type=store_true) = False,
):
    inputs = i
    out_path = Path(o)
    out_path.mkdir(exist_ok=True)

    files = flatten([get_files(p, recurse=True) for p in inputs])

    for p in inputs:
        if not Path(p).exists(): raise NotADirectoryError(f"Invalid path: {p}")

    before_moving = get_files(out_path)
    for f in tqdm(files):
        if move: shutil.move(f, out_path / f.name)
        else:    shutil.copy(f, out_path / f.name)
    after_moving = get_files(out_path)
    new_files = set(after_moving) - set(before_moving)

    logger.info(f"Of {len(files)} files from input folders, moved {len(new_files)} files to output folder")



from upyog.ml.imports import *


__all__ = ["IMAGENET_MEAN", "IMAGENET_STDEV", "PreprocessingParams"]


IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STDEV = (0.229, 0.224, 0.225)


@dataclass
class PreprocessingParams:
    input_height: int
    input_width: int
    normalisation_mean: Optional[List[float]]
    normalisation_stdev: Optional[List[float]]
    interpolation: str
    output_format: Literal["BGR", "RGB"]
    resize_method: Literal["squish", "pad"]
    pad_fill: Optional[Tuple[int, int, int]] = None

    @property
    def _pil_interpolation_map():
        return {
             "bicubic" : Image.Resampling.BICUBIC,
            "bilinear" : Image.Resampling.BILINEAR,
             "lanczos" : Image.Resampling.LANCZOS,
             "nearest" : Image.Resampling.NEAREST,
                 "box" : Image.Resampling.BOX,
             "hamming" : Image.Resampling.HAMMING,
        }

    @staticmethod
    def _check_stats_in_0_to_1_range(stats: Tuple[float, float, float]):
        for x in stats:
            if x > 1.0:
                raise RuntimeError(f"Normalisation stats should be between 0-1. Got {stats}")

    @property
    def pil_interpolation_method(self) -> Image.Resampling:
        return self._pil_interpolation_map[self.interpolation]

    def __post_init__(self):
        if self.normalisation_mean is not None:  self._check_stats_in_0_to_1_range(self.normalisation_mean)
        if self.normalisation_stdev is not None: self._check_stats_in_0_to_1_range(self.normalisation_stdev)
        if not self.output_format in ("BGR", "RGB"):          raise ValueError
        if not self.interpolation in self._pil_interpolation_map.keys():
            raise ValueError(
                f"resize method '{self.interpolation}' is invalid. Valid choices are: {list(self._pil_interpolation_map.keys())}"
            )

    def create_coreml_input(self, input_name: str = "Image") -> List['ct.ImageType']:
        from upyog.ml.coreml_utils import get_coreml_input

        return get_coreml_input(
            H = self.input_height,
            W = self.input_width,
            norm_mean = self.normalisation_mean,
            norm_stdev = self.normalisation_stdev,
            input_name = input_name,
            input_format = self.output_format,
        )



from upyog.imports import *

# Inspired by https://github.com/gordicaleksa/stable_diffusion_playground/blob/a4508d0650c6940f64b858aa818291cfa0bdad9c/generate_images.py#L84-L97
def save_image_with_metadata(
    img: Image.Image, metadata: dict, filepath: PathLike
):
    """
    Embed `metadata` inside an image as the Exif UserComment when saving it to disk
    """

    import exif
    import cv2

    filepath = Path(filepath)
    ext = filepath.suffix

    # If no extension given, add .jpg extension
    if ext == "":
        filepath = filepath.parent / f"{filepath.name}.jpg"

    elif ext != "jpg":
        raise ValueError(f"Only '.jpg' extensions are acceptable, you provided {ext}")

    assert json.dumps(metadata), f"JSON Data not serialisable!"
    assert ".jpg" in filepath, f"Expected full filepath with '.jpg' extension in filename"

    # See https://exif.readthedocs.io/en/latest/usage.html
    exif_img = exif.Image(
        cv2.imencode(".jpg", np.asarray(img)[..., ::-1])[1].tobytes()
    )
    exif_img.user_comment = json.dumps(metadata)

    with open(filepath, "wb") as f:
        f.write(exif_img.get_file())


def fig_to_pil(fig, pad_inches: float = 0.25):
    """
    Convert a Matplotlib figure to a PIL Image and return it
    """
    # Save figure to buf in png format
    buf = io.BytesIO()
    fig.savefig(buf, format='png', bbox_inches='tight', pad_inches=pad_inches)
    buf.seek(0)

    # Convert to PIL Image
    img = Image.open(buf).convert("RGB")
    buf.close()
    return img



from upyog.imports import *
from itertools import islice

__all__ = [
    "flatten", "uniqueify",
    "chunk", "notnone", "lmap", "allequal", "zipsafe", "convert_size", "convert_to_tuple",
    "convert_to_list",
]


def flatten(x: Any) -> List[Any]:
    flattened_list = []
    for item in x:
        if hasattr(item, "__iter__") and not isinstance(item, (str, bytes)):
            [flattened_list.append(i) for i in item]
        else:
            flattened_list.append(item)
    return flattened_list


def convert_to_list(x: Union[str, Iterable[str]]):
    # fmt: off
    if   isinstance(x, list):       return x
    elif isinstance(x, np.ndarray): return x.tolist()
    elif isinstance(x, tuple):      return list(x)
    elif isinstance(x, str):        return [x]
    else: raise TypeError(f"Expected {str}, got {type(x)}")
    # fmt: on


def convert_to_tuple(x: Union[str, Iterable[str]]) -> Tuple[str]:
    # fmt: off
    if   isinstance(x, list):              return tuple(x)
    elif isinstance(x, np.ndarray):        return tuple(x)
    elif isinstance(x, tuple):             return x
    elif isinstance(x, (str, int, float)): return x,
    else: raise TypeError(f"Expected {{str|int|float|list|np.ndarray|tuple}}, got {type(x)}")
    # fmt: on


def uniqueify(x: Collection) -> Collection:
    return sorted(list(set(x)))


def sort_dict_by_keys(x) -> dict:
    return dict(sorted(x.items()))


def get_uuid(n=10) -> str:
    return uuid.uuid4().hex[:n].lower()


# Borrowed from https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
def chunk(items: Collection[Any], size: int, strict=True) -> list:
    if strict:
        assert len(items) % size == 0

    items = iter(items)
    return list(iter(lambda: tuple(islice(items, size)), ()))


# Borrowed from icevision.utils
def notnone(x):
    return x is not None


def lmap(f, xs):
    return list(map(f, xs)) if notnone(xs) else None


def allequal(l):
    return l.count(l[0]) == len(l) if l else True


def zipsafe(*its):
    if not allequal(lmap(len, its)):
        raise ValueError("The elements have different leghts")
    return zip(*its)


# https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
def convert_size(size_bytes: int) -> str:
    "Convert bytes into appropriate unit and return a human readable string for file size"
    if size_bytes == 0:
       return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])



from upyog.imports import *
from upyog.image.draw import *
from upyog.image.visualiser import Visualiser


__all__ = [
    "draw_rule_of_thirds",
    "img_join_horizontal",
    "img_join_vertical",
    "make_img_grid",
]


ImageCollection = Collection[Image.Image]


def draw_rule_of_thirds(img: Image.Image, thickness=5, opacity=0.4) -> Image.Image:
    vis = Visualiser(img)
    vis.draw_rule_of_thirds(thickness, opacity)
    return vis.img


def img_join_horizontal(imgs: ImageCollection) -> Image.Image:
    return reduce(operator.__or__, imgs)


def img_join_vertical(imgs: ImageCollection) -> Image.Image:
    return reduce(operator.__floordiv__, imgs)


def make_img_grid(
    imgs: ImageCollection,
    ncol=3,
    size_WH: Optional[tuple] = (640, 384),
    pad=True,
    verbose=False,
):
    # fmt: off
    if size_WH:
        if pad: imgs = [img.resize_pad(size_WH) for img in imgs]
        else:   imgs = [img.resize(size_WH) for img in imgs]
    else:
        pass
    # fmt: on

    if ncol == 1:
        return img_join_vertical(imgs)

    if verbose:
        from rich.progress import Progress

        with Progress() as prog:
            # Create rows with individual images
            rows, row = [], []
            for i, img in enumerate(imgs, start=1):
                row += [img]
                if i % ncol == 0:
                    rows += [row]
                    row = []
            if not row == []:
                rows += [row]

            n_tasks = len(rows) + 1
            task_concat = prog.add_task("Concatenating Rows...", total=n_tasks)
            # print(task_concat.total)
            grid = []

            while not prog.finished:
                for row in rows:
                    grid += [img_join_horizontal(row)]
                    prog.update(task_concat, advance=1)

                grid = img_join_vertical(grid)
                prog.update(task_concat, advance=1)

    else:
        rows, row = [], []
        for i, img in enumerate(imgs, start=1):
            row += [img]
            if i % ncol == 0:
                rows += [row]
                row = []
        if not row == []:
            rows += [row]

        grid = [img_join_horizontal(row) for row in rows]
        grid = img_join_vertical(grid)

    return grid



from upyog.imports import *

def zip_folder(path: PathLike):
    path = Path(path)
    assert path.is_dir()
    assert path.exists()

    return shutil.make_archive(path, "zip", path)

def compose_files_into_zip_archive(
    files: List[PathLike],
    out_dir: PathLike,
    zip_name: str = "Archive",
):
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True)
    zip_name = zip_name + ".zip"

    for file in files:
        shutil.copy(file, str(out_dir / zip_name))

    zipfile = zip_folder(out_dir)
    shutil.rmtree(out_dir)

    return zipfile



from upyog.imports import *
from upyog.os.read_files import PathLike, get_files

__all__ = [
    "load_json", "read_json", "check_pil_simd_usage", "sanitise_filename", "get_file_size",
    "get_file_creation_date", "write_json", "write_text",
    "is_platform_macos", "is_platform_windows", "is_platform_linux",
    "convert_number_to_human_readable_format",
]


def load_json(path: PathLike):
    "Load a JSON file"
    return json.loads(Path(path).read_bytes())


def read_json(path: PathLike):
    return load_json(path)


def write_json(json_data: dict, path: PathLike, indent=4):
    "Write `json_data` to `path`"
    with open(path, "w") as f:
        json.dump(json_data, f, indent=indent)


def write_text(text: str, save_path: PathLike):
    "Write text to a text file"
    with open(str(save_path), 'w') as f:
        f.write(text)


def check_pil_simd_usage():
    pil_version = PIL.__version__

    if not "post" in pil_version:
        msg = f"Pillow-SIMD not installed. Using PIL {pil_version} instead"
        warnings.warn(msg)

    else:
        if not PILFeatures.check_feature("libjpeg_turbo"):
            warnings.warn(
                f"Pillow-SIMD is installed, but Libjpeg Turbo is not being used. "
                f"Unlikely to see any speedups"
            )


def check_corrupted_images(filepaths: List[PathLike], verbose=True) -> List[PathLike]:
    corrupted = []
    for f in tqdm(filepaths, "Scanning files for corruptions", disable=not verbose):
        try:    Image.open(f)
        except: corrupted.append(f)

    if verbose:
        logger.info(f"Found {len(corrupted)} corrupt files")

    return corrupted


# fmt: off
def sanitise_filename(
    f: str, lowercase=False, prefix=None, truncate: Optional[int] = 240
):
    # see https://stackoverflow.com/questions/49460802/regex-match-all-vular-fractions
    fractions = "\u00BC-\u00BE\u2150-\u215E"  # not-exhaustive..?
    supscripts = "\u00B1-\u00B9"

    fn = re.sub(f'[\W{supscripts}{fractions}]', '_', f) # captures (?) subscripts, fractions, other non-alphanumerics
    fn = re.sub(f'[^A-Za-z0-9_+]', '_', fn)             # captures alphabets in foreign languages
    fn = re.sub('^[\W_]*'     , '' , fn)                # replace leading spl characters or '_'

    assert re.match("^[A-Za-z0-9_]*$", fn), f'{fn} contains non-alphanumeric/underscore characters'
    assert re.match("^[A-Za-z0-9]",    fn), f'{fn} OR {f} starts with a non-alphanumeric character'

    if prefix: fn = f"{prefix}{fn}"
    if truncate: fn = fn[:truncate]
    if lowercase: fn = fn.lower()

    return fn


def get_file_size(path: PathLike, units: Literal["B", "KB", "MB", "GB"] = "MB"):
    path = Path(path)
    if path.is_dir():
        B = sum(os.path.getsize(f) for f in get_files(path, recurse=True))
    else:
        B = os.path.getsize(path)

    if   units == "B":  return B
    elif units == "KB": return B/1024
    elif units == "MB": return B/1024**2
    elif units == "GB": return B/1024**3


def get_file_creation_date(path_to_file) -> str:
    # Get the timestamp of the file's creation date
    timestamp = os.path.getctime(path_to_file)
    
    # Convert the timestamp to a datetime object
    date = datetime.fromtimestamp(timestamp)
    return date.strftime('%Y-%m-%d')


def convert_number_to_human_readable_format(number: Union[int, float], approx=True) -> str:
    """
    Formats a number to be human readable. Works best for larger numbers up to
    1 quintillion (1e18)
              100 -> 100
            10000 -> 10k
          100,000 -> 100k
      100,000,000 -> 100m
    1,000,000,000 -> 1b
    """
    magnitude = 0
    while abs(number) >= 1000:
        magnitude += 1
        number /= 1000.0

    if approx:
        return '%d%s' % (number, ['', 'k', 'm', 'b', 't', 'q'][magnitude])
    else:
        return '%.1f%s' % (number, ['', 'k', 'm', 'b', 't', 'q'][magnitude])


def is_platform_macos() -> bool:   return platform.system().lower() == "darwin"
def is_platform_windows() -> bool: return platform.system().lower() == "windows"
def is_platform_linux() -> bool:   return platform.system().lower() == "linux"



from upyog.ml.imports import *
from upyog.ml.utils import to_np, to_torch
from upyog.image import fig_to_pil


__all__ = ["visualise_conf_matrix"]


class MetricsVisualiser:
    def __init__(
        self,
        ground_truths: Union[Collection[float], Collection[str]],
        predictions: Union[Collection[float], Collection[str]],
    ):
        self.ground_truths = to_np(ground_truths, dtype=None)
        self.predictions = to_np(predictions, dtype=None)

    @classmethod
    def from_df(
        cls,
        df: pd.DataFrame, ground_truths_colname: str, predictions_colname: str, 
    ):
        return cls(
            ground_truths = df[ground_truths_colname].values.tolist(),
            predictions = df[predictions_colname].values.tolist(),
        )

    def visualise_confusion_matrix():
        ...

    def visualise_metrics():
        ...

    def get_classification_report():
        ...


def get_confusion_matrix(
    ground_truths: Union[List[float], List[str]],
    predictions: Union[List[float], List[str]],
):
    from sklearn.metrics import confusion_matrix

    if not len(ground_truths) == len(predictions):
        raise RuntimeError(f"Mismatch between `len(ground_truths)` and `len(predictions)`")
    return confusion_matrix(ground_truths, predictions)

def visualise_conf_matrix(
    conf_matrix: np.ndarray,
    labels: List[str],
    title: Optional[str] = None,
    normalize: bool = True,
    cmap: Union[None, str, mpl.colors.Colormap] = None,
    width: int = 6,
    dpi: int = 200,
    return_pil: bool = True,
):
    # Convert to DataFrame
    cm = conf_matrix
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        cm = pd.DataFrame(cm * 100)
    else:
        cm = pd.DataFrame(cm)
    cm.index = labels
    cm.columns = labels

    # Setup plot canvas
    fig, ax = plt.subplots(figsize=(width, width), dpi=dpi)

    # Colormap
    if not cmap:
        cmap = 'Blues'
        # cmap = sns.cubehelix_palette(as_cmap=True)

    # Draw heatmap, axes and title
    sns.heatmap(cm, annot=True, fmt=".1f", cbar=False, ax=ax, cmap=cmap)
    _= ax.set_yticklabels(labels, rotation=0)
    _= ax.set_xticklabels(labels, rotation=90)

    if title:
        ax.set_title(title)

    if return_pil: return fig_to_pil(fig)
    else:          return fig



from upyog.imports import *
import pytz


__all__ = [
    "get_YYYY_MM_DD", "get_date_YYYY_MM_DD", "get_date_DD_MM_YYYY",
    "get_current_time_pst", "get_current_time_ist", "get_current_time_utc",
]


def get_YYYY_MM_DD(sep="_") -> str:
    warnings.warn(f"`get_YYYY_MM_DD` is deprecated. Please call `get_date_YYYY_MM_DD()` instead")
    return get_date_YYYY_MM_DD()


def get_date_YYYY_MM_DD(sep="_") -> str:
    return datetime.now().strftime(f"%Y{sep}%m{sep}%d")


def get_date_DD_MM_YYYY(sep="_") -> str:
    return datetime.now().strftime(f"%d{sep}%m{sep}%Y")


_TIME_FORMAT_STRING = "%Y/%m/%d at %H:%M:%S"

def _utc_localized_time_now():
    return pytz.utc.localize(datetime.utcnow())


def get_current_time_pst(fmt_str: Optional[str] = _TIME_FORMAT_STRING):
    utc_now = _utc_localized_time_now()
    pst_now = utc_now.astimezone(pytz.timezone("America/Los_Angeles"))

    return pst_now.strftime(fmt_str) if fmt_str else pst_now


def get_current_time_ist(fmt_str: Optional[str] = _TIME_FORMAT_STRING):
    utc_now = _utc_localized_time_now()
    ist_now = utc_now.astimezone(pytz.timezone("Asia/Kolkata"))

    return ist_now.strftime(fmt_str) if fmt_str else ist_now


def get_current_time_utc(fmt_str: Optional[str] = _TIME_FORMAT_STRING):
    utc_now = _utc_localized_time_now()
    return utc_now.strftime(fmt_str) if fmt_str else utc_now


# TODO
# def get_cest_time_now(fmt_str):
#     ...

# def get_pst_time_now(fmt_str):
#     ...

# def get_time_by_timezone(timezone: str, fmt_str: Optional[str]):
#     # switch by timezone
#     ...



from upyog.imports import *
from torch.utils.data import Dataset, DataLoader
from torch import Tensor
from torchvision.io.image import ImageReadMode

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torchvision.transforms as Tfms
import torchvision.transforms.functional as TF

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns


IS_COREML_AVAILABLE = is_package_available("coremltools")
if IS_COREML_AVAILABLE:
    import coremltools as ct



def xyxy_to_xywh(xyxy):
    assert len(xyxy) == 4
    x1, y1, x2, y2 = xyxy
    w = x2 - x1
    h = y2 - y1
    return (x1, y1, w, h)


def xyxys_to_xywh(xyxys):
    return [xyxy_to_xywh(xyxy) for xyxy in xyxys]


def xywh_to_xyxy(xywh):
    assert len(xywh) == 4
    x1, y1, w, h = xywh
    x2 = x1 + w
    y2 = y1 + h
    return (x1, y1, x2, y2)


def xywhs_to_xyxy(xywhs):
    return [xywh_to_xyxy(xywh) for xywh in xywhs]



from upyog.imports import *

__all__ = ["load_image", "resize_with_padding"]


def load_image(fn, mode: Optional[str] = "RGB") -> Image.Image:
    img = Image.open(fn)
    img.load()
    img = img._new(img.im)
    return img.convert(mode) if mode else img


@fastcore.patch
def resize_pad(
    self: Image.Image,
    size_WH: tuple,
    pad_location: Literal["center", "top"] = "center",
    fill: int = 0,
):
    return resize_with_padding(self, size_WH, fill, pad_location)


def get_ratio(from_WH, to_WH) -> np.ndarray:
    W, H = from_WH
    to_W, to_H = to_WH

    return min([to_H / H, to_W / W])


def resize_with_padding(
    img: Image.Image,
    target_WH,
    pad_fill: Union[int, Tuple[int, int, int]] = 0,
    pad_location: Literal["center", "top"] = "center",
    # mode: str = "RGB",
):
    to_W, to_H = target_WH
    W, H = img.size
    orig_mode = img.mode

    ratio = get_ratio(img.size, target_WH)
    W, H = int(ratio * W), int(ratio * H)

    if pad_location == "top":
        box = (0, 0)
    else:
        box = ((to_W - W) // 2, (to_H - H) // 2)

    # Create a new image and paste the resized on it
    if pad_fill:
        if   isinstance(pad_fill, int):          color = (pad_fill, pad_fill, pad_fill)
        elif isinstance(pad_fill, (list,tuple)): color = tuple(pad_fill)
        assert isinstance(color, tuple)
    else:
        color = None

    padded_img = Image.new(orig_mode, (to_W, to_H), color=color)
    padded_img.paste(img.resize((W, H), Image.ANTIALIAS), box)

    return padded_img
